-------------------------------------------------------------------------------
--
-- Audio Controller OPB Peripheral
--
--
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity opb_audio_controller is
  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"FEFF0400";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"FEFF04FF");

  port (
    -- OPB Input Signals
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address

    -- OPB Output Signals
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge

    -- Audio IO Signals
    AU_BUSY     : out std_logic;        -- Audio interface busy
    AU_CSN      : out std_logic;        -- Audio Chip Select 
    AU_CCLK     : out std_logic;        -- Audio Cntl Clock
    AU_CDTI     : out std_logic;        -- Audio Cntl Data In  
    PB_DIN      : in  std_logic_vector(0 to 15)
    --AU_CDTO     : in  std_logic	    -- Audio Cntl Data Out 
    );       
end opb_audio_controller;


architecture behavioral of opb_audio_controller is
    
  signal AU_CDTO: std_logic;
  signal Command : std_logic_vector(0 to 15);
  signal Status: std_logic_vector (0 to 7);
  signal Busy : std_logic;
  signal SetBusy : std_logic;
  signal RNW : std_logic;     
  signal CmdLoad: std_logic;
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  signal audio_done: std_logic;
  signal shift_enable: std_logic;
  signal capture_enable: std_logic;
  signal shift_mode: std_logic;
  signal rising: std_logic;
  signal falling: std_logic;
  signal clear: std_logic;  
  signal cs_setup: std_logic;
  signal cs_hold: std_logic;
  signal baud_counter: std_logic_vector ( 0 to 3 );
  signal bit_counter: std_logic_vector (0 to 3);
  signal bit_counter_ov: std_logic;
  
  constant ST_AU_IDLE:  std_logic_vector( 0 to 2) := "000";
  constant ST_AU_START:  std_logic_vector( 0 to 2) := "001";
  constant ST_AU_SETUP:  std_logic_vector( 0 to 2) := "010";
  constant ST_AU_SHIFT:  std_logic_vector( 0 to 2) := "011";
  constant ST_AU_END:  std_logic_vector( 0 to 2) := "100";
  constant ST_AU_HOLD:  std_logic_vector( 0 to 2) := "101";
  signal au_current_state, au_next_state : std_logic_vector (0 to 2);

  signal xfer_start: std_logic;    -- Transfer start
  signal xfer_end: std_logic;      -- Transfer end     
  signal xfer_idle: std_logic;     -- No transfer pending
  signal obuf_oe: std_logic;       -- Data output buffer enable
  signal xfer_done: std_logic;     -- Transfer has been done

  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant ST_IDLE      : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant ST_SELECTED  : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant ST_WAIT      : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant ST_XFER      : std_logic_vector(0 to STATE_BITS-1) := "111";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin
  
  ----------------------------------------------------------------------------
  -- OPB interface
  ----------------------------------------------------------------------------
  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      SetBusy <=  '0';
      RNW <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      SetBusy <= OPB_DBus(0);
      RNW <= OPB_RNW;
    end if;
  end process register_opb_inputs;

  register_opb_outputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Sln_DBus <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
     if (output_enable = '1' and RNW = '1')  then
        Sln_DBus <= Busy & "0000000" & Status & "0000000000000000";
     else
       Sln_DBus <= (others => '0');
     end if;
    end if;
  end process register_opb_outputs;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-1) = C_BASEADDR(0 to C_OPB_AWIDTH-1) else
    '0';

  Sln_xferAck <= present_state(0);

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= ST_IDLE;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(present_state, chip_select, OPB_Select)
  begin   
      case present_state is
        when ST_IDLE =>
          if chip_select = '1' then
            next_state <= ST_SELECTED;
          else
            next_state <= ST_IDLE;
          end if;
        when ST_SELECTED =>
          if OPB_Select = '1' then
              next_state <= ST_WAIT;
          else
            next_state <= ST_IDLE;
          end if;
        when ST_WAIT =>
          if OPB_Select = '1' then
            next_state <= ST_XFER;
          else
            next_state <= ST_IDLE;
          end if;
        -- State encoding is critical here: xfer must only be true here
        when ST_XFER =>
          next_state <= ST_IDLE;
        when others =>
          next_state <= ST_IDLE;
      end case;

  end process fsm_comb;

  process (present_state) 
  begin
     if ( present_state = ST_WAIT) then 
         output_enable <= '1';
     else 
         output_enable <= '0';
     end if;
  end process;
  
  -----------------------------------------------------------------------------
  -- Audio interface
  ----------------------------------------------------------------------------

  AU_CDTO <= PB_DIN (13); -- PB_D2
   
  process (present_state , SetBusy) 
  begin
     if ( present_state = ST_WAIT and SetBusy = '1' and RNW = '0' ) then 
         CmdLoad <= '1';
     else 
        CmdLoad <= '0';
     end if;
  end process;
  
  -- Audio controller is busy shifting out the command
  -- Must be assigned the PH bus at this time
  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        Busy <= '0';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( CmdLoad = '1' ) then 
           Busy <= '1';
        elsif (audio_done = '1') then 
           Busy <= '0';
        end if;
     end if;
  end process;
  AU_BUSY <= Busy;
  
  -- Audio command register
  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        Command <= (others => '0');
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( CmdLoad = '1' ) then 
           Command <= OPB_DBus(16 to 31);
        elsif (shift_enable = '1') then 
           Command <= Command( 1 to 15) & '0';
        end if;
     end if;
  end process;
  
  -- Audio controll data output 
  AU_CDTI <= Command (0);
  
  -- Audio read back data
  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        Status <= (others => '0');
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if (capture_enable = '1') then 
           Status <= AU_CDTO & Status( 0 to 6) ;
        end if;
     end if;
  end process;

  -- Counter for the CCLK divider 
  -- 50MHz / 16
  process (OPB_Clk, OPB_Rst) 
  begin
     if ( OPB_Rst = '1') then 
        baud_counter <= "0000";
     elsif ( OPB_Clk 'event and OPB_Clk = '1') then 
        if (clear = '1') then 
           baud_counter <= "0000";
        else 
           baud_counter <= baud_counter + '1'; 
        end if;
     end if;
  end process;
  
  -- Detect the time for the CCLK rising and falling
  process ( baud_counter ) 
  begin
     if ( baud_counter = "0111" ) then 
        rising <= '1'; 
     else 
        rising <= '0';
     end if;
     
     if ( baud_counter = "1111" ) then 
        falling <= '1'; 
     else 
        falling <= '0';
     end if;

  end process;
  
  -- Audio controller serial shift bit counter
  process ( OPB_Clk, OPB_Rst) 
  begin
     if (OPB_Rst = '1' ) then 
        bit_counter <= "0000";
     elsif ( OPB_Clk 'event and OPB_Clk = '1') then 
        if ( clear = '1' or cs_setup = '1' ) then 
           bit_counter <= "0000";
        elsif (falling = '1') then 
           bit_counter <= bit_counter + '1';
        end if;
     end if;
  end process;
  
  process (bit_counter ) 
  begin
     if ( bit_counter = "1111") then 
        bit_counter_ov <= '1';
     else 
        bit_counter_ov <= '0';
     end if;
  end process;
  
  -- Data output shift enable 
  shift_enable <= shift_mode and falling;   
  
  -- Data input shift enable
  capture_enable <= shift_mode and rising;
    
  -- Audio controller state machine
  process ( OPB_Clk, OPB_Rst) 
  begin
     if ( OPB_Rst = '1') then 
        au_current_state <= ST_AU_IDLE;
     elsif (OPB_Clk 'event and OPB_Clk = '1') then 
        au_current_state <= au_next_state;
     end if;
  end process;
  
  -- Audio controller next state logic
  process ( au_current_state, CmdLoad, rising, bit_counter_ov) 
  begin
     case au_current_state is 
        when ST_AU_IDLE =>
           if ( CmdLoad = '1') then  
              au_next_state <= ST_AU_START;
           else 
              au_next_state <= ST_AU_IDLE;
           end if;
        when ST_AU_START =>
           if ( rising = '1' ) then 
              au_next_state <= ST_AU_SETUP;
           else 
              au_next_state <= ST_AU_START;
           end if;
        when ST_AU_SETUP =>
           if ( rising = '1' ) then 
              au_next_state <= ST_AU_SHIFT;
           else 
              au_next_state <= ST_AU_SETUP;
           end if;
        when ST_AU_SHIFT =>
           if ( rising = '1' and bit_counter_ov = '1') then 
              au_next_state <= ST_AU_HOLD; 
           else 
              au_next_state <= ST_AU_SHIFT;
           end if;
        when ST_AU_HOLD => 
           if ( rising = '1' ) then 
              au_next_state <= ST_AU_END;
           else 
              au_next_state <= ST_AU_HOLD;
           end if;               
        when ST_AU_END =>
           au_next_state <= ST_AU_IDLE;
        when others => 
           au_next_state <= ST_AU_IDLE;
     end case;
  
  end process;
  
  -- Audio controller control signal decode from state machine
  process ( au_current_state, rising, bit_counter_ov) 
  begin         
     shift_mode <= '0'; 
     clear <= '0';  
     cs_setup <= '0';  
     cs_hold <= '0';
     audio_done <= '0';
     case au_current_state is 
        when ST_AU_IDLE =>  
           shift_mode <= '0'; 
           clear <= '1';  
           cs_setup <= '0';
           audio_done <= '0';
        when ST_AU_SETUP =>
           cs_setup <= '1';
        when ST_AU_SHIFT =>  
           shift_mode <= '1';
        when ST_AU_HOLD => 
           cs_hold <= '1';
        when ST_AU_END =>
           audio_done <= '1'; 
        when others =>  
           shift_mode <= '0'; 
           clear <= '0';  
           cs_setup <= '0';
           audio_done <= '0';
           cs_hold <= '0';
     end case;
  end process;

  process ( OPB_Clk, OPB_Rst) 
  begin
     if ( OPB_Rst = '1') then 
        AU_CSN <= '1';
     elsif (OPB_Clk 'event and OPB_Clk = '1') then 
        if ( clear = '1' or audio_done = '1') then 
           AU_CSN <= '1';
        elsif ( cs_setup = '1' ) then 
           AU_CSN <= '0';
        end if ;
     end if;
  end process;
  
  process ( OPB_Clk, OPB_Rst) 
  begin
     if ( OPB_Rst = '1') then 
        AU_CCLK <= '1';
     elsif (OPB_Clk 'event and OPB_Clk = '1') then 
        if ( clear = '1' or cs_hold = '1' or rising = '1') then 
           AU_CCLK <= '1';
        elsif ( falling = '1' and cs_hold = '0' ) then 
           AU_CCLK <= '0';
        end if ;
     end if;
  end process;
   
end behavioral;
