------------------------------
--
------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity opb_ethernet is
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"00800000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"00FFFFFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge

    PB_DIN   : in std_logic_vector(0 to 15);      -- ETHER data bus
    PB_DOUT  : out std_logic_vector(0 to 15);     -- ETHER data bus output
    PB_DOUT_EN : out std_logic;                   -- ETHER data bus output enable
    PB_A : out std_logic_vector (0 to 19);
    IORD_N : out std_logic; 
    IOWR_N : out std_logic;
    AEN : out std_logic;
    BHE_N : out std_logic;
    ETHERNET_CS_N : out std_logic;
    ETHERNET_RDY : in std_logic;
    ETHERNET_IREQ : in std_logic;
    ETHERNET_IOCS16_N : in std_logic    
	 );
end opb_ethernet;

architecture Behavioral of opb_ethernet is


  constant ETHER_AWIDTH : integer := 10;  -- Number of address lines on the ETHER
  constant ETHER_DWIDTH : integer := 16;  -- Number of data lines on the ETHER
    
  signal RNW : std_logic;
  signal ETHER_DI: std_logic_vector(0 to 31);
  signal ABus : std_logic_vector(0 to ETHER_AWIDTH-1);
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  signal TE : std_logic;
  signal BE : std_logic_vector(0 to 3);
  signal A30: std_logic;
  signal ETHER_RD_DATA : std_logic_vector (0 to 31);
  signal ETHER_WR_DATA : std_logic_vector (0 to 31);
  
  signal CE : std_logic;           -- ETHER chip enalbe
  signal OE : std_logic;           -- ETHER output enalbe
  signal xfer_start: std_logic;    -- Transfer start
  signal xfer_end: std_logic;      -- Transfer end     
  signal xfer_idle: std_logic;     -- No transfer pending
  signal obuf_oe: std_logic;       -- Data output buffer enable
  signal xfer_done: std_logic;     -- Transfer has been done
  signal data_latch: std_logic;   

  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant ST_IDLE      : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant ST_SELECTED  : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant ST_WAIT      : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant ST_XFER      : std_logic_vector(0 to STATE_BITS-1) := "111";
  constant ST_WAIT1     : std_logic_vector(0 to STATE_BITS-1) := "100";
  constant ST_WAIT2     : std_logic_vector(0 to STATE_BITS-1) := "101";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin

  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      ETHER_DI <= (others => '0');
      ABus <= (others => '0');
      RNW <= '0';
      A30 <=  '0';
      BE <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
      ETHER_DI <= OPB_DBus(0 to C_OPB_DWIDTH-1);
      ABus <= OPB_ABus(C_OPB_AWIDTH-2-(ETHER_AWIDTH-1) to C_OPB_AWIDTH-2);
      RNW <= OPB_RNW;
      A30 <= OPB_ABus(30);
      BE <= OPB_BE(0 to 3);
    end if;
  end process register_opb_inputs;

  register_opb_outputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Sln_DBus <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
--     if ( output_enable = '1' and RNW = '1') then
     if ( output_enable = '1') then
        Sln_DBus <= ETHER_RD_DATA;
     else
       Sln_DBus <= (others => '0');
     end if;
    end if;
  end process register_opb_outputs;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';
  --Sln_DBus(ETHER_DWIDTH to C_OPB_DWIDTH-1) <= (others => '0');

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-2-ETHER_AWIDTH) =
        C_BASEADDR(0 to C_OPB_AWIDTH-2-ETHER_AWIDTH) else
    '0';

  Sln_xferAck <= xfer_done;

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= ST_IDLE;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(present_state, chip_select, OPB_Select)
  begin   
      case present_state is
        when ST_IDLE =>
          if chip_select = '1' then
            next_state <= ST_SELECTED;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_SELECTED =>
          if OPB_Select = '1' then
              next_state <= ST_WAIT;
          else
            next_state <= ST_IDLE;
          end if;
   
        when ST_WAIT =>
          if OPB_Select = '1' then
            next_state <= ST_WAIT1;
          else
            next_state <= ST_IDLE;
          end if;
        when ST_WAIT1 =>
            if OPB_Select = '1' then
              next_state <= ST_WAIT2;
            else
              next_state <= ST_IDLE;
            end if;
        when ST_WAIT2 =>
              if OPB_Select = '1' then
                next_state <= ST_XFER;
              else
                next_state <= ST_IDLE;
              end if;
   
        -- State encoding is critical here: xfer must only be true here
        when ST_XFER =>
          next_state <= ST_IDLE;
          
        when others =>
          next_state <= ST_IDLE;
      end case;

  end process fsm_comb;

  fsm_output: process (present_state) 
  begin
     output_enable <= '0';
     xfer_start <= '0';
     xfer_end <= '0';   
     xfer_idle <= '0';
     xfer_done <= '0';
     case present_state is  
        when ST_IDLE =>
           xfer_idle <= '1';
        when ST_SELECTED =>
           xfer_start <= '1';
        when ST_WAIT2 =>      
           output_enable <= '1';
        when ST_XFER =>        
           xfer_end <= '1';
           xfer_done <= '1';
        when others => 
           output_enable <= '0';
           xfer_start <= '0';
           xfer_end <= '0';
           xfer_idle <= '0';
     end case;
  end process fsm_output;
  
  -- The following process generates the OE output signal
  ether_oe_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        OE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( chip_select = '1' and xfer_idle = '1') then 
           OE <= not OPB_RNW;
        elsif ( xfer_end = '1') then 
           OE <= '1';
        end if;
     end if;
  end process ether_oe_reg;

  -- The following process generates the chip enable output signal
  ether_ce_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        CE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( chip_select = '1' and xfer_idle = '1') then 
           CE <= '0';
        elsif ( xfer_end = '1') then 
           CE <= '1';
        end if;
     end if;
  end process ether_ce_reg;                              
  
  -- The following process generates the TE output signal
  ether_we_reg: process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        TE <= '1';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( xfer_start = '1') then 
           TE <= '0';
        elsif (output_enable = '1') then
           TE <= '1';
        end if;
     end if;
  end process ether_we_reg;                              
  
  --
  
  ETHER_RD_DATA <= PB_DIN & PB_DIN;
  
 --  ETHER_WR_DATA <= ETHER_DI;
  process ( ETHER_DI, A30)
  begin
        if ( A30 = '1') then 
           ETHER_WR_DATA (16 to 31) <= ETHER_DI(16 to 31);
        else 
           ETHER_WR_DATA (16 to 31) <= ETHER_DI(0 to 15);
        end if;
  end process;

  IORD_N <= OE;
  
  -- The following assign the output signals
  process ( RNW, TE) 
  begin
     if ( RNW = '1') then 
--        IORD_N <= TE;
        IOWR_N <= '1';
     else 
--        IORD_N <= '1';
        IOWR_N <= TE;
     end if;
  end process;
  
--  process ( BE ) 
--  begin 
--     if ( BE = "1100" or BE = "0011" ) then 
--        BHE_N <= '0';
--     else 
--        BHE_N <= '1';
--     end if;
--  end process; 

  BHE_N <= CE;
  
  ETHERNET_CS_N <= CE;
  AEN <= CE;
  PB_A <= "0000000000" & ABus;
  PB_DOUT <= ETHER_WR_DATA(16 to 31);
  PB_DOUT_EN <= (not CE) and OE;


end Behavioral;
