-------------------------------------------------------------------------------
--
-- OPB Bus Multiplexer
--
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity opb_pb_mux is
  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"FFEF0500";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"FFFF05FF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge
    
    PB_D   : inout std_logic_vector(0 to 15);      -- SRAM data bus
    PB_A   : out std_logic_vector (0 to 19);      -- SRAM address bus
    PB_OE_N : out std_logic;                      -- SRAM output enable
    PB_WE_N : out std_logic;                      -- SRAM write enable
    PB_UB_N : out std_logic;                      -- SRAM byte enable
    PB_LB_N : out std_logic;                      -- SRAM byte enable 
    PB_DIN  : out std_logic_vector (0 to 15);     -- SRAM data bus


    PB_DOUT_SRAM   : in std_logic_vector(0 to 15);
    PB_DOUT_EN_SRAM   : in std_logic;
    PB_A_SRAM   : in std_logic_vector(0 to 19);
    PB_OE_N_SRAM   : in std_logic;
    PB_WE_N_SRAM   : in std_logic;
    PB_UB_N_SRAM   : in std_logic;
    PB_LB_N_SRAM   : in std_logic;

    PB_DOUT_FLASH   : in std_logic_vector(0 to 15);
    PB_DOUT_EN_FLASH   : in std_logic;
    PB_A_FLASH   : in std_logic_vector(0 to 19);
    PB_OE_N_FLASH   : in std_logic;
    PB_WE_N_FLASH   : in std_logic;

    PB_DOUT_SDRAM   : in std_logic_vector(0 to 15);
    PB_DOUT_EN_SDRAM   : in std_logic;
    PB_A_SDRAM   : in std_logic_vector(0 to 19);
    PB_UB_N_SDRAM   : in std_logic;
    PB_LB_N_SDRAM   : in std_logic;

    PB_DOUT_ETHER   : in std_logic_vector(0 to 15);
    PB_DOUT_EN_ETHER   : in std_logic;
    PB_A_ETHER   : in std_logic_vector(0 to 19);
    PB_OE_N_ETHER   : in std_logic;
    PB_WE_N_ETHER   : in std_logic;
    PB_UB_N_ETHER   : in std_logic;
    PB_LB_N_ETHER   : in std_logic;
    
    AU_BUSY   : in std_logic;
    AU_CCLK   : in std_logic;
    AU_CDTI   : in std_logic;
    --AU_CDTO   : out std_logic;

    SDRAM_LOCK : out std_logic;
    SDRAM_ACTIVE : in std_logic
    );       

end opb_pb_mux;

architecture Behavioral of opb_pb_mux is

  component IOBUF_F_24
    port(
      O  : out   std_ulogic;
      IO : inout std_ulogic;
      I  : in    std_ulogic;
      T  : in    std_ulogic
      );
  end component;
  
  ----- component OBUF_F_24               -----
  component OBUF_F_24
    port(
      O : out std_ulogic;
      I : in  std_ulogic
      );
  end component;
  
  
  signal current_slave : std_logic_vector (0 to 2) ;  
  signal Abus : std_logic_vector (0 to 8);
  signal PB_DOUT : std_logic_vector(0 to 15); 
  signal PB_DOUT_EN : std_logic;  

  signal SetLock : std_logic;
  signal Lock : std_logic;
  signal RNW : std_logic;
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  
  signal xfer_start: std_logic;    -- Transfer start
  signal xfer_end: std_logic;      -- Transfer end     
  signal xfer_idle: std_logic;     -- No transfer pending
  signal obuf_oe: std_logic;       -- Data output buffer enable
  signal xfer_done: std_logic;     -- Transfer has been done
  signal oe_audio: std_logic;
  
  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant ST_IDLE      : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant ST_SELECTED  : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant ST_WAIT      : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant ST_XFER      : std_logic_vector(0 to STATE_BITS-1) := "111";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin

  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      SetLock <=  '0';
      RNW <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      SetLock <= OPB_DBus(C_OPB_DWIDTH-1);
      RNW <= OPB_RNW;
    end if;
  end process register_opb_inputs;

  register_opb_outputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Sln_DBus <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
     if (output_enable = '1' and RNW = '1')  then
        Sln_DBus <= "0000000000000000" & "00000000" & "000000" & Lock & '0';
     else
       Sln_DBus <= (others => '0');
     end if;
    end if;
  end process register_opb_outputs;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-1) = C_BASEADDR(0 to C_OPB_AWIDTH-1) else
    '0';

  Sln_xferAck <= present_state(0);

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= ST_IDLE;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(present_state, chip_select, OPB_Select)
  begin   
      case present_state is
        when ST_IDLE =>
          if chip_select = '1' then
            next_state <= ST_SELECTED;
          else
            next_state <= ST_IDLE;
          end if;
        when ST_SELECTED =>
          if OPB_Select = '1' then
              next_state <= ST_WAIT;
          else
            next_state <= ST_IDLE;
          end if;
        when ST_WAIT =>
          if OPB_Select = '1' then
            next_state <= ST_XFER;
          else
            next_state <= ST_IDLE;
          end if;
        -- State encoding is critical here: xfer must only be true here
        when ST_XFER =>
          next_state <= ST_IDLE;
        when others =>
          next_state <= ST_IDLE;
      end case;

  end process fsm_comb;

  process (present_state) 
  begin
     if ( present_state = ST_WAIT) then 
         output_enable <= '1';
     else 
         output_enable <= '0';
     end if;
  end process;
  
  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        Lock <= '0';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( present_state = ST_WAIT and SetLock = '1' ) then 
           Lock <= '1';
        elsif (SDRAM_ACTIVE = '1') then 
           Lock <= '0';
        end if;
     end if;
  end process;

  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        SDRAM_LOCK <= '0';
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( Lock = '1' ) then 
           SDRAM_LOCK <= '1';
        end if;
     end if;
  end process;


--  PB_DIN <= PB_D;

--  process (PB_DOUT, PB_DOUT_EN) begin
--     if ( PB_DOUT_EN = '1') then
--        PB_D <= PB_DOUT;
--     else
--        PB_D <= "ZZZZZZZZZZZZZZZZ";
--     end if;
--  end process;
  
  Abus <= OPB_ABus ( 0 to 8);
  
  process (OPB_Clk, OPB_Rst) 
  begin 
     if OPB_Rst = '1' then 
        current_slave <= "000";
     elsif OPB_Clk 'event and OPB_Clk = '1' then 
        if ( Lock = '1') then 
            current_slave <= "010";
        elsif ( AU_BUSY = '1') then 
            current_slave <= "100";
        elsif ( OPB_select = '1') then 
           case Abus is 
              when "001000000" =>  -- 20000000
                  current_slave <= "000";
              when "001100000" =>  -- 30000000
                  current_slave <= "001";
              when "010000000" =>  -- 40000000
                  current_slave <= "010";
              when "000000001" =>  -- 00800000
                  current_slave <= "011";
              when "111111111" =>  -- FFEF0000 
                  current_slave <= "100";
              when others => 
                  current_slave <= "010";
           end case;  
        end if;
     end if;
  end process;

  --AU_CDTO <= PB_D(13);
  --AU_CDTO <= '0';
  
  process (     PB_DOUT_SRAM,  
                PB_DOUT_EN_SRAM,
                PB_A_SRAM, 
                PB_OE_N_SRAM,  
                PB_WE_N_SRAM,  
                PB_UB_N_SRAM,  
                PB_LB_N_SRAM,  
            
                PB_DOUT_FLASH, 
                PB_DOUT_EN_FLASH, 
                PB_A_FLASH, 
                PB_OE_N_FLASH, 
                PB_WE_N_FLASH, 
            
                PB_DOUT_SDRAM, 
                PB_DOUT_EN_SDRAM, 
                PB_A_SDRAM,  
                PB_UB_N_SDRAM,  
                PB_LB_N_SDRAM,  
            
                PB_DOUT_ETHER, 
                PB_DOUT_EN_ETHER, 
                PB_A_ETHER,  
                PB_OE_N_ETHER, 
                PB_WE_N_ETHER, 
                PB_UB_N_ETHER, 
                PB_LB_N_ETHER, 
            
                AU_CCLK,  
                AU_CDTI   ) 
  begin
     case current_slave is 
        when "000" => 
          PB_A       <= PB_A_SRAM; 
          PB_OE_N    <= PB_OE_N_SRAM;
          PB_WE_N    <= PB_WE_N_SRAM;
          PB_UB_N    <= PB_UB_N_SRAM;
          PB_LB_N    <= PB_LB_N_SRAM;
          PB_DOUT    <= PB_DOUT_SRAM;
          PB_DOUT_EN <= PB_DOUT_EN_SRAM;        
        when "001" =>
          PB_A       <= PB_A_FLASH; 
          PB_OE_N    <= PB_OE_N_FLASH;
          PB_WE_N    <= PB_WE_N_FLASH;
          PB_UB_N    <= '1';
          PB_LB_N    <= '1';
          PB_DOUT    <= PB_DOUT_FLASH;
          PB_DOUT_EN <= PB_DOUT_EN_FLASH;
        when "010" => 
          PB_A       <= PB_A_SDRAM; 
          PB_OE_N    <= '1';
          PB_WE_N    <= '1';
          PB_UB_N    <= PB_UB_N_SDRAM;
          PB_LB_N    <= PB_LB_N_SDRAM;
          PB_DOUT    <= PB_DOUT_SDRAM;
          PB_DOUT_EN <= PB_DOUT_EN_SDRAM;
        when "011" =>
          PB_A       <= PB_A_ETHER; 
          PB_OE_N    <= PB_OE_N_ETHER;
          PB_WE_N    <= PB_WE_N_ETHER;
          PB_UB_N    <= PB_UB_N_ETHER;
          PB_LB_N    <= PB_LB_N_ETHER;
          PB_DOUT    <= PB_DOUT_ETHER;
          PB_DOUT_EN <= PB_DOUT_EN_ETHER;
                
        when "100" => 
          PB_A       <= (others => '0') ; 
          PB_OE_N    <= '1';
          PB_WE_N    <= '1';
          PB_UB_N    <= '1';
          PB_LB_N    <= '1';
          PB_DOUT    <= "00000000000000" & AU_CDTI & AU_CCLK;
          PB_DOUT_EN <= '1';
                
        when others =>
          PB_A       <= PB_A_SDRAM; 
          PB_OE_N    <= '1';
          PB_WE_N    <= '1';
          PB_UB_N    <= PB_UB_N_SDRAM;
          PB_LB_N    <= PB_LB_N_SDRAM;
          PB_DOUT    <= PB_DOUT_SDRAM;
          PB_DOUT_EN <= PB_DOUT_EN_SDRAM;
     end case;
  end process;


  obuf_oe <= not PB_DOUT_EN;  
  
  -- Audio CDTO must be input mode in case of audio controller transfer
  process ( PB_DOUT_EN , AU_BUSY) 
  begin
     if ( AU_BUSY = '1') then 
        oe_audio <= '1';
     else 
        oe_audio <= not PB_DOUT_EN; 
     end if;
  end process;
  
  -- The following assign the data pads
  data0_pad: IOBUF_F_24 port map (
  O => PB_DIN(0),
  IO => PB_D(0),
  I => PB_DOUT(0),
  T => obuf_oe
  );
  
  data1_pad: IOBUF_F_24 port map (
  O => PB_DIN(1),
  IO => PB_D(1),
  I => PB_DOUT(1),
  T => obuf_oe
  );
  
  data2_pad: IOBUF_F_24 port map (
  O => PB_DIN(2),
  IO => PB_D(2),
  I => PB_DOUT(2),
  T => obuf_oe
  );
  
  data3_pad: IOBUF_F_24 port map (
  O => PB_DIN(3),
  IO => PB_D(3),
  I => PB_DOUT(3),
  T => obuf_oe
  );
  
  data4_pad: IOBUF_F_24 port map (
  O => PB_DIN(4),
  IO => PB_D(4),
  I => PB_DOUT(4),
  T => obuf_oe
  );
  
  data5_pad: IOBUF_F_24 port map (
  O => PB_DIN(5),
  IO => PB_D(5),
  I => PB_DOUT(5),
  T => obuf_oe
  );
  
  data6_pad: IOBUF_F_24 port map (
  O => PB_DIN(6),
  IO => PB_D(6),
  I => PB_DOUT(6),
  T => obuf_oe
  );
  
  data7_pad: IOBUF_F_24 port map (
  O => PB_DIN(7),
  IO => PB_D(7),
  I => PB_DOUT(7),
  T => obuf_oe
  );
  
  data8_pad: IOBUF_F_24 port map (
  O => PB_DIN(8),
  IO => PB_D(8),
  I => PB_DOUT(8),
  T => obuf_oe
  );
  
  data9_pad: IOBUF_F_24 port map (
  O => PB_DIN(9),
  IO => PB_D(9),
  I => PB_DOUT(9),
  T => obuf_oe
  );
  
  data10_pad: IOBUF_F_24 port map (
  O => PB_DIN(10),
  IO => PB_D(10),
  I => PB_DOUT(10),
  T => obuf_oe
  );
  
  data11_pad: IOBUF_F_24 port map (
  O => PB_DIN(11),
  IO => PB_D(11),
  I => PB_DOUT(11),
  T => obuf_oe
  );
  
  data12_pad: IOBUF_F_24 port map (
  O => PB_DIN(12),
  IO => PB_D(12),
  I => PB_DOUT(12),
  T => obuf_oe
  );
  
  data13_pad: IOBUF_F_24 port map (
  O => PB_DIN(13),
  IO => PB_D(13),
  I => PB_DOUT(13),
  T => oe_audio
  );
  
  data14_pad: IOBUF_F_24 port map (
  O => PB_DIN(14),
  IO => PB_D(14),
  I => PB_DOUT(14),
  T => obuf_oe
  );
  
  data15_pad: IOBUF_F_24 port map (
  O => PB_DIN(15),
  IO => PB_D(15),
  I => PB_DOUT(15),
  T => obuf_oe
  );
  
end Behavioral;
